import json
import os
from datetime import datetime

import pytest
from freezegun import freeze_time
from senor_octopus.sources.weatherapi import weatherapi

dirname, filename = os.path.split(os.path.abspath(__file__))
with open(os.path.join(dirname, "weatherapi_response.json")) as fp:
    mock_payload = json.load(fp)


@freeze_time("2021-01-01")
@pytest.mark.asyncio
async def test_weatherapi(mocker, httpx_mock) -> None:
    mocker.patch.dict(
        os.environ,
        {"WEATHERAPI_TOKEN": "XXX"},
    )
    httpx_mock.add_response(json=mock_payload)

    events = [event async for event in weatherapi("iceland")]
    assert events == [
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.last_updated_epoch",
            "value": 1616644811,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.last_updated",
            "value": "2021-03-25 04:00",
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.temp_c",
            "value": -2.0,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.temp_f",
            "value": 28.4,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.is_day",
            "value": 0,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.condition.text",
            "value": "Partly cloudy",
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.condition.icon",
            "value": "//cdn.weatherapi.com/weather/64x64/night/116.png",
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.condition.code",
            "value": 1003,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.wind_mph",
            "value": 2.5,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.wind_kph",
            "value": 4.0,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.wind_degree",
            "value": 100,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.wind_dir",
            "value": "E",
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.pressure_mb",
            "value": 988.0,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.pressure_in",
            "value": 29.6,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.precip_mm",
            "value": 0.0,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.precip_in",
            "value": 0.0,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.humidity",
            "value": 93,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.cloud",
            "value": 75,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.feelslike_c",
            "value": -6.5,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.feelslike_f",
            "value": 20.4,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.vis_km",
            "value": 10.0,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.vis_miles",
            "value": 6.0,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.uv",
            "value": 1.0,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.gust_mph",
            "value": 11.0,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.current.gust_kph",
            "value": 17.6,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.forecast.forecastday.maxtemp_c",
            "value": 3.7,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.forecast.forecastday.maxtemp_f",
            "value": 38.7,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.forecast.forecastday.mintemp_c",
            "value": 0.2,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.forecast.forecastday.mintemp_f",
            "value": 32.4,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.forecast.forecastday.avgtemp_c",
            "value": 1.7,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.forecast.forecastday.avgtemp_f",
            "value": 35.1,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.forecast.forecastday.maxwind_mph",
            "value": 10.7,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.forecast.forecastday.maxwind_kph",
            "value": 17.3,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.forecast.forecastday.totalprecip_mm",
            "value": 1.7,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.forecast.forecastday.totalprecip_in",
            "value": 0.07,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.forecast.forecastday.avgvis_km",
            "value": 9.0,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.forecast.forecastday.avgvis_miles",
            "value": 5.0,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.forecast.forecastday.avghumidity",
            "value": 77.0,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.forecast.forecastday.daily_will_it_rain",
            "value": 1,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.forecast.forecastday.daily_chance_of_rain",
            "value": "74",
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.forecast.forecastday.daily_will_it_snow",
            "value": 0,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.forecast.forecastday.daily_chance_of_snow",
            "value": "0",
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.forecast.forecastday.condition.text",
            "value": "Patchy rain possible",
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.forecast.forecastday.condition.icon",
            "value": "//cdn.weatherapi.com/weather/64x64/day/176.png",
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.forecast.forecastday.condition.code",
            "value": 1063,
        },
        {
            "timestamp": datetime(2021, 1, 1),
            "name": "hub.weatherapi.forecast.forecastday.uv",
            "value": 1.0,
        },
    ]



import logging
import os
from datetime import datetime

import httpx
from senor_octopus.lib import flatten
from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def weatherapi(location: str, prefix: str = "hub.weatherapi") -> Stream:
    token = os.environ["WEATHERAPI_TOKEN"]

    _logger.info("Fetching weather data")
    url = (
        f"http://api.weatherapi.com/v1/forecast.json?key={token}"
        f"&q={location}&days=2&aqi=no&alerts=no"
    )
    async with httpx.AsyncClient() as client:
        response = await client.get(url)
    payload = response.json()
    _logger.debug("Received %s", payload)

    for key, value in flatten(payload["current"]).items():
        yield {
            "timestamp": datetime.utcnow(),
            "name": f"{prefix}.current.{key}",
            "value": value,
        }

    tomorrow = payload["forecast"]["forecastday"][1]
    for key, value in flatten(tomorrow["day"]).items():
        yield {
            "timestamp": datetime.utcnow(),
            "name": f"{prefix}.forecast.forecastday.{key}",
            "value": value,
        }



"""
    Dummy conftest.py for senor_octopus.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    - https://docs.pytest.org/en/stable/fixture.html
    - https://docs.pytest.org/en/stable/writing_plugins.html
"""
import pytest
from senor_octopus.cli import CaseConfigParser
from senor_octopus.graph import build_dag


config_content = """
[random]
plugin = source.random
flow = -> check, normal
schedule = * * * * *

[check]
plugin = filter.jsonpath
flow = random -> high
filter = $.events[?(@.value>0.5)]

[normal]
plugin = sink.log
flow = random ->

[high]
plugin = sink.log
flow = check ->
throttle = 5 minutes
level = warning
"""


@pytest.fixture
def mock_config():
    config = CaseConfigParser()
    config.read_string(config_content)
    yield config


@pytest.fixture
async def mock_dag(mock_config):
    dag = build_dag(mock_config)
    yield dag



import asyncio
import logging
from typing import Dict
from typing import Set

from senor_octopus.graph import Source

_logger = logging.getLogger(__name__)


SLEEP_TIME = 5


class Scheduler:
    def __init__(self, dag: Set[Source]):
        self.dag = dag

    async def run(self) -> None:
        nodes = {node for node in self.dag if node.schedule}
        if not nodes:
            _logger.info("Nothing to schedule")
            return

        _logger.info("Starting scheduler")
        loop = asyncio.get_event_loop()

        schedules: Dict[str, float] = {}
        while True:
            for node in nodes:
                now = loop.time()
                delay = node.schedule.next(default_utc=False)
                when = now + delay

                if node.name not in schedules:
                    _logger.info(f"Scheduling {node.name} to run in {delay} seconds")
                    schedules[node.name] = when
                elif schedules[node.name] <= now:
                    _logger.info(f"Running {node.name}")
                    await node.run()
                    del schedules[node.name]

            _logger.debug(f"Sleeping for {SLEEP_TIME} seconds")
            await asyncio.sleep(SLEEP_TIME)



import logging
import os
import textwrap

import aiopg
from psycopg2 import sql
from psycopg2.extras import Json
from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def postgresql(stream: Stream, table: str = "events") -> None:
    dbname = os.environ["POSTGRES_DBNAME"]
    host = os.environ["POSTGRES_HOST"]
    port = os.environ["POSTGRES_PORT"]
    user = os.environ["POSTGRES_USER"]
    password = os.environ["POSTGRES_PASSWORD"]

    dsn = f"dbname={dbname} user={user} password={password} host={host} port={port}"
    async with aiopg.create_pool(dsn) as pool:
        async with pool.acquire() as conn:
            async with conn.cursor() as cur:
                _logger.debug("Trying to create table `%s`", table)
                await cur.execute(
                    sql.SQL(
                        textwrap.dedent(
                            """
                            CREATE TABLE IF NOT EXISTS {table} (
                                "timestamp" TIMESTAMP,
                                "name" VARCHAR,
                                "value" JSON
                            );
                            """,
                        ),
                    ).format(table=sql.Identifier(table)),
                )
                _logger.debug("Trying to create index `%s_name_idx`", table)
                await cur.execute(
                    sql.SQL(
                        textwrap.dedent(
                            """
                            CREATE INDEX IF NOT EXISTS {index}
                            ON {table} USING HASH(name);
                            """,
                        ),
                    ).format(
                        table=sql.Identifier(table),
                        index=sql.Identifier(f"{table}_name_idx"),
                    ),
                )

                _logger.info("Inserting events into Postgres")
                async for event in stream:  # pragma: no cover
                    _logger.debug(event)
                    await cur.execute(
                        sql.SQL(
                            textwrap.dedent(
                                """
                                INSERT INTO {table} ("timestamp", "name", "value")
                                VALUES (%s, %s, %s);
                                """,
                            ),
                        ).format(table=sql.Identifier(table)),
                        (event["timestamp"], event["name"], Json(event["value"])),
                    )



import asyncio
import random
from typing import Dict
from unittest import mock

import aiotools
import pytest
from freezegun import freeze_time
from senor_octopus.cli import CaseConfigParser
from senor_octopus.graph import build_dag
from senor_octopus.graph import connected
from senor_octopus.graph import Source


def test_connected() -> None:
    config = {
        "a": {"flow": "-> *"},
        "b": {"flow": "-> c"},
        "c": {"flow": "* -> d, e"},
        "d": {"flow": "* ->"},
        "e": {"flow": "c, b ->"},
    }
    assert not connected(config, "a", "b")
    assert connected(config, "a", "c")
    assert connected(config, "a", "d")
    assert not connected(config, "a", "e")
    assert not connected(config, "b", "a")
    assert connected(config, "b", "c")
    assert not connected(config, "b", "d")
    assert not connected(config, "b", "e")
    assert not connected(config, "c", "a")
    assert not connected(config, "c", "b")
    assert connected(config, "c", "d")
    assert connected(config, "c", "e")
    assert not connected(config, "d", "a")
    assert not connected(config, "d", "b")
    assert not connected(config, "d", "c")
    assert not connected(config, "d", "e")
    assert not connected(config, "e", "a")
    assert not connected(config, "e", "b")
    assert not connected(config, "e", "c")
    assert not connected(config, "e", "d")


@pytest.mark.asyncio
async def test_build_dag(mock_config) -> None:
    dag = build_dag(mock_config)
    assert len(dag) == 1

    source = dag.pop()
    assert isinstance(source, Source)
    assert len(source.next) == 2


@pytest.mark.asyncio
async def test_build_dag_many_to_one() -> None:
    config = CaseConfigParser()
    config.read_string(
        """
        [one]
        plugin = source.random
        flow = -> three

        [two]
        plugin = source.random
        flow = -> three

        [three]
        plugin = sink.log
        flow = * ->
    """,
    )
    dag = build_dag(config)
    assert len(dag) == 2

    one, two = dag
    assert one.next == two.next


def test_build_dag_missing_plugin() -> None:
    config = CaseConfigParser()
    config.read_string(
        """
        [a]
        flow = -> *
    """,
    )
    with pytest.raises(Exception) as excinfo:
        build_dag(config)
    assert str(excinfo.value) == "Invalid config, missing `plugin` key"


@pytest.mark.asyncio
async def test_build_dag_seen() -> None:
    config = CaseConfigParser()
    config.read_string(
        """
        [a]
        flow = -> *
        plugin = source.random

        [b]
        flow = -> *
        plugin = source.random

        [c]
        flow = * ->
        plugin = sink.log
    """,
    )
    build_dag(config)


def test_build_dag_missing_flow() -> None:
    config = CaseConfigParser()
    config.read_string(
        """
        [a]
        plugin = source.random
    """,
    )
    with pytest.raises(Exception) as excinfo:
        build_dag(config)
    assert str(excinfo.value) == "Invalid config, missing `flow` key"


def test_build_dag_invalid_plugin() -> None:
    config = CaseConfigParser()
    config.read_string(
        """
        [a]
        plugin = source.invalid
        flow = -> *
    """,
    )
    with pytest.raises(Exception) as excinfo:
        build_dag(config)
    assert str(excinfo.value) == "Invalid plugin name `source.invalid`"


def test_build_dag_environ(mocker) -> None:
    mock_env: Dict[str, str] = {}
    mocker.patch("senor_octopus.graph.os.environ", mock_env)

    config = CaseConfigParser()
    config.read_string(
        """
        [a]
        flow = -> *
        plugin = source.random
        A_ENV_VAR = 1
    """,
    )
    build_dag(config)
    assert mock_env == {"A_ENV_VAR": "1"}


@freeze_time("2021-01-01")
@pytest.mark.asyncio
async def test_run_source(mocker, mock_config) -> None:
    mock_logger = mock.MagicMock()
    mocker.patch("senor_octopus.sinks.log._logger", mock_logger)
    random.seed(42)

    dag = build_dag(mock_config)
    source = dag.pop()

    await source.run()
    assert len(mock_logger.log.mock_calls) == 14

    # test throttle, should get 10 entries from the non-throttled sink
    await source.run()
    assert len(mock_logger.log.mock_calls) == 24


@pytest.mark.asyncio
async def test_batch(mocker) -> None:
    mock_logger = mock.MagicMock()
    mocker.patch("senor_octopus.sinks.log._logger", mock_logger)
    vclock = aiotools.VirtualClock()

    config_content = """
[random]
plugin = source.random
flow = -> log
schedule = * * * * *

[log]
plugin = sink.log
flow = random ->
batch = 2 minutes
    """
    config = CaseConfigParser()
    config.read_string(config_content)

    with vclock.patch_loop():
        dag = build_dag(config)
        source = dag.pop()

        await source.run()
        assert len(mock_logger.log.mock_calls) == 0

        await asyncio.sleep(180)
        assert len(mock_logger.log.mock_calls) == 10


@pytest.mark.asyncio
async def test_batch_empty_source(mocker) -> None:
    mock_logger = mock.MagicMock()
    mocker.patch("senor_octopus.sinks.log._logger", mock_logger)
    vclock = aiotools.VirtualClock()

    config_content = """
[random]
plugin = source.random
flow = -> log
schedule = * * * * *
events = 0

[log]
plugin = sink.log
flow = random ->
batch = 2 minutes
    """
    config = CaseConfigParser()
    config.read_string(config_content)

    with vclock.patch_loop():
        dag = build_dag(config)
        source = dag.pop()

        await source.run()
        assert len(mock_logger.log.mock_calls) == 0

        await asyncio.sleep(180)
        assert len(mock_logger.log.mock_calls) == 0



from unittest import mock

import aiotools
import pytest
from asynctest import CoroutineMock
from senor_octopus.scheduler import Scheduler


@pytest.mark.asyncio
async def test_scheduler() -> None:
    mock_source1 = mock.MagicMock()
    mock_source1.schedule.next.return_value = 10
    mock_source1.run = CoroutineMock()
    mock_source1.run.side_effect = [None, None, Exception("Stopped")]
    mock_source2 = mock.MagicMock()
    mock_source2.schedule = None
    mock_dag = {mock_source1, mock_source2}
    vclock = aiotools.VirtualClock()

    with vclock.patch_loop():
        scheduler = Scheduler(mock_dag)  # type: ignore
        with pytest.raises(Exception) as excinfo:
            await scheduler.run()

    assert str(excinfo.value) == "Stopped"
    assert len(mock_source1.run.mock_calls) == 3


@pytest.mark.asyncio
async def test_scheduler_no_jobs() -> None:
    vclock = aiotools.VirtualClock()
    with vclock.patch_loop():
        scheduler = Scheduler({})  # type: ignore
        await scheduler.run()



import logging
from datetime import datetime

import speedtest
from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def speed(prefix: str = "hub.speedtest") -> Stream:
    _logger.info("Testing internet speed")
    s = speedtest.Speedtest()
    s.get_best_server()
    s.download()
    s.upload()
    _logger.debug("Received %s", s.results.dict())

    for key, value in s.results.dict().items():
        yield {
            "timestamp": datetime.utcnow(),
            "name": f"{prefix}.{key}",
            "value": value,
        }



import os
import random

import pytest
from senor_octopus.sinks.pushover import pushover
from senor_octopus.sources.rand import rand


@pytest.mark.asyncio
async def test_pushover(mocker, httpx_mock) -> None:
    mocker.patch.dict(
        os.environ,
        {
            "PUSHOVER_APP_TOKEN": "XXX",
            "PUSHOVER_USER_TOKEN": "alice",
        },
    )
    httpx_mock.add_response()
    random.seed(42)

    await pushover(rand(2))
    requests = httpx_mock.get_requests()
    assert len(requests) == 2
    assert (
        requests[0].read()
        == b"token=XXX&user=alice&message=hub.random%3A+0.6394267984578837"
    )


@pytest.mark.asyncio
async def test_pushover_empty_stream(mocker, httpx_mock) -> None:
    mocker.patch.dict(
        os.environ,
        {
            "PUSHOVER_APP_TOKEN": "XXX",
            "PUSHOVER_USER_TOKEN": "alice",
        },
    )
    random.seed(42)

    await pushover(rand(0))
    assert not httpx_mock.get_request()



import random
from datetime import datetime
from typing import Union

from senor_octopus.types import Stream


async def rand(events: Union[str, int] = 10) -> Stream:
    for _ in range(int(events)):
        yield {
            "timestamp": datetime.now(),
            "name": "hub.random",
            "value": random.random(),
        }



import os
import random
from datetime import datetime
from unittest import mock

import pytest
from asynctest import CoroutineMock
from freezegun import freeze_time
from psycopg2 import sql
from senor_octopus.sinks.db.postgresql import postgresql
from senor_octopus.sources.rand import rand


@freeze_time("2021-01-01")
@pytest.mark.asyncio
async def test_postgresql(mocker) -> None:
    mocker.patch.dict(
        os.environ,
        {
            "POSTGRES_DBNAME": "dbname",
            "POSTGRES_HOST": "localhost",
            "POSTGRES_PORT": "5432",
            "POSTGRES_USER": "user",
            "POSTGRES_PASSWORD": "XXX",
        },
    )
    mock_aiopg = CoroutineMock()
    mock_cursor = (
        mock_aiopg.create_pool.return_value.__aenter__.return_value.acquire.return_value.__aenter__.return_value.cursor.return_value.__aenter__.return_value
    )
    mock_cursor.execute = CoroutineMock()
    mocker.patch("senor_octopus.sinks.db.postgresql.aiopg", mock_aiopg)
    random.seed(42)

    await postgresql(rand(1))

    mock_cursor.execute.assert_has_calls(
        [
            mock.call(
                sql.Composed(
                    [
                        sql.SQL("\nCREATE TABLE IF NOT EXISTS "),
                        sql.Identifier("events"),
                        sql.SQL(
                            ' (\n    "timestamp" TIMESTAMP,\n    "name" VARCHAR,\n    "value" JSON\n);\n',
                        ),
                    ],
                ),
            ),
            mock.call(
                sql.Composed(
                    [
                        sql.SQL("\nCREATE INDEX IF NOT EXISTS "),
                        sql.Identifier("events_name_idx"),
                        sql.SQL("\nON "),
                        sql.Identifier("events"),
                        sql.SQL(" USING HASH(name);\n"),
                    ],
                ),
            ),
            mock.call(
                sql.Composed(
                    [
                        sql.SQL("\nINSERT INTO "),
                        sql.Identifier("events"),
                        sql.SQL(
                            ' ("timestamp", "name", "value")\nVALUES (%s, %s, %s);\n',
                        ),
                    ],
                ),
                (
                    datetime(2021, 1, 1, 0, 0),
                    "hub.random",
                    mock.ANY,  # Json(0.6394267984578837)
                ),
            ),
        ],
    )


@pytest.mark.asyncio
async def test_postgresql_empty_stream(mocker) -> None:
    mocker.patch.dict(
        os.environ,
        {
            "POSTGRES_DBNAME": "dbname",
            "POSTGRES_HOST": "localhost",
            "POSTGRES_PORT": "5432",
            "POSTGRES_USER": "user",
            "POSTGRES_PASSWORD": "XXX",
        },
    )
    mock_aiopg = CoroutineMock()
    mock_cursor = (
        mock_aiopg.create_pool.return_value.__aenter__.return_value.acquire.return_value.__aenter__.return_value.cursor.return_value.__aenter__.return_value
    )
    mock_cursor.execute = CoroutineMock()
    mocker.patch("senor_octopus.sinks.db.postgresql.aiopg", mock_aiopg)
    random.seed(42)

    await postgresql(rand(0))
    assert len(mock_cursor.execute.mock_calls) == 2



import logging
import sys
from unittest import mock

import pytest
from asynctest import CoroutineMock
from senor_octopus.cli import main
from senor_octopus.cli import parse_args
from senor_octopus.cli import run
from senor_octopus.cli import setup_logging


def test_parse_args() -> None:
    with pytest.raises(SystemExit) as excinfo:
        parse_args(["--version"])
    assert str(excinfo.value) == "0"

    parser = parse_args(["config.ini", "-vv"])
    assert parser.f == "config.ini"
    assert parser.loglevel == logging.DEBUG

    parser = parse_args(["config.ini", "-v"])
    assert parser.f == "config.ini"
    assert parser.loglevel == logging.INFO


def test_setup_logging(mocker, capfd) -> None:
    mock_logging = mock.MagicMock()
    mocker.patch("senor_octopus.cli.logging", mock_logging)

    setup_logging(logging.WARNING)

    mock_logging.basicConfig.assert_called_with(
        level=logging.WARNING,
        stream=sys.stdout,
        format="[%(asctime)s] %(levelname)s:%(name)s:%(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )


@pytest.mark.asyncio
async def test_main(mocker) -> None:
    mocker.patch("senor_octopus.cli.CaseConfigParser")
    mocker.patch("senor_octopus.cli.build_dag")

    mock_scheduler = mock.MagicMock()
    mock_scheduler.return_value.run = CoroutineMock()
    mock_scheduler.return_value.run.side_effect = [None, Exception("Stopped")]
    mocker.patch("senor_octopus.cli.Scheduler", mock_scheduler)

    await main(["config.ini"])

    mock_scheduler.return_value.run.assert_called()


def test_run(mocker) -> None:
    mock_main = CoroutineMock()
    mocker.patch("senor_octopus.cli.main", mock_main)
    mocker.patch("senor_octopus.cli.sys.argv", ["srocto", "config.ini", "-vv"])

    run()

    mock_main.assert_called_with(["config.ini", "-vv"])



# flake8: noqa
import pytest
from senor_octopus.cli import CaseConfigParser
from senor_octopus.graph import build_dag
from senor_octopus.lib import flatten
from senor_octopus.lib import render_dag


def test_flatten() -> None:
    assert flatten({"a": 1, "b": 2.0, "c": "foo", "d": {"e": "bar"}}) == {
        "a": 1,
        "b": 2.0,
        "c": "foo",
        "d.e": "bar",
    }


def test_render_dag(mock_dag) -> None:
    assert (
        render_dag(mock_dag, use_color=False).strip()
        == r"""
*   random
|\  
* | check
| * normal
* high
        """.strip()
    )


@pytest.mark.asyncio
async def test_render_dag_many_to_many() -> None:
    config = CaseConfigParser()
    config.read_string(
        """
        [one]
        plugin = source.random
        flow = -> two, three

        [two]
        plugin = filter.jsonpath
        flow = one -> four
        filter = $.events[?(@.value>0.5)]

        [three]
        plugin = filter.jsonpath
        flow = one -> four
        filter = $.events[?(@.value<=0.5)]

        [four]
        plugin = sink.log
        flow = two, three ->
        """,
    )

    dag = build_dag(config)
    assert (
        render_dag(dag, use_color=False).strip()
        == r"""
*   one
|\  
* | three
| * two
|/  
* four
        """.strip()
    )



import argparse
import asyncio
import configparser
import logging
import sys

from senor_octopus import __version__
from senor_octopus.graph import build_dag
from senor_octopus.lib import render_dag
from senor_octopus.scheduler import Scheduler

__author__ = "Beto Dealmeida"
__copyright__ = "Beto Dealmeida"
__license__ = "MIT"

_logger = logging.getLogger(__name__)


def parse_args(args):
    """Parse command line parameters

    Args:
      args (List[str]): command line parameters as list of strings
          (for example  ``["--help"]``).

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """
    parser = argparse.ArgumentParser(description="Schedule pipelines")
    parser.add_argument(
        "--version",
        action="version",
        version="senor-octopus {ver}".format(ver=__version__),
    )
    parser.add_argument(
        dest="f",
        help="Location of the config file",
        type=str,
        metavar="PATH",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO,
    )
    parser.add_argument(
        "-vv",
        "--very-verbose",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG,
    )
    return parser.parse_args(args)


def setup_logging(loglevel):
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(
        level=loglevel,
        stream=sys.stdout,
        format=logformat,
        datefmt="%Y-%m-%d %H:%M:%S",
    )


class CaseConfigParser(configparser.RawConfigParser):
    def optionxform(self, optionstr):
        return optionstr


async def main(args):
    args = parse_args(args)
    setup_logging(args.loglevel)

    _logger.info("Reading configuration")
    config = CaseConfigParser()
    config.read(args.f)
    _logger.info("Building DAG")
    dag = build_dag(config)
    _logger.info("\n%s", render_dag(dag))

    _logger.info("Running Sr. Octopus")
    await Scheduler(dag).run()

    _logger.info("Done")


def run():
    asyncio.run(main(sys.argv[1:]))


if __name__ == "__main__":
    run()



import logging
import os

import httpx
from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def pushover(stream: Stream) -> None:
    app_token = os.environ["PUSHOVER_APP_TOKEN"]
    user_token = os.environ["PUSHOVER_USER_TOKEN"]

    url = "https://api.pushover.net/1/messages.json"
    async for event in stream:  # pragma: no cover
        _logger.debug(event)
        data = {
            "token": app_token,
            "user": user_token,
            "message": "{name}: {value}".format(**event),
        }
        _logger.info("Posting message to Pushover")
        async with httpx.AsyncClient() as client:
            await client.post(url, data=data)



import logging

from jsonpath import JSONPath
from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def jsonpath(stream: Stream, filter: str) -> Stream:
    """
    Filter event stream based on a JSON path.
    """
    _logger.debug("Filtering events")
    events = [event async for event in stream]
    for event in JSONPath(filter).parse({"events": events}):
        yield event



import logging

from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def log(stream: Stream, level: str = "INFO") -> None:
    async for event in stream:
        _logger.log(getattr(logging, level.upper()), event)



from io import StringIO
from typing import Any
from typing import Dict
from typing import List
from typing import Set
from typing import Tuple
from typing import Union

import asciidag.graph
import asciidag.node
from senor_octopus.graph import Node
from senor_octopus.graph import Source


def flatten(
    obj: Dict[str, Any],
    prefix: str = "",
    sep: str = ".",
) -> Dict[str, Union[str, int, float]]:
    items: List[Tuple[str, Union[str, int, float]]] = []
    for key, value in obj.items():
        new_key = sep.join((prefix, key)) if prefix else key
        if isinstance(value, dict):
            items.extend(flatten(value, new_key, sep=sep).items())
        else:
            items.append((new_key, value))
    return dict(items)


def render_dag(dag: Set[Source], **kwargs: Any) -> str:
    out = StringIO()
    graph = asciidag.graph.Graph(out, **kwargs)
    asciidag_nodes: Dict[str, asciidag.node.Node] = {}
    tips = sorted(
        [build_asciidag(node, asciidag_nodes) for node in dag],
        key=lambda n: n.item,
    )

    graph.show_nodes(tips)
    out.seek(0)
    return out.getvalue()


def build_asciidag(
    node: Node,
    asciidag_nodes: Dict[str, asciidag.node.Node],
) -> asciidag.node.Node:
    if node.name in asciidag_nodes:
        asciidag_node = asciidag_nodes[node.name]
    else:
        asciidag_node = asciidag.node.Node(node.name)
        asciidag_nodes[node.name] = asciidag_node

    asciidag_node.parents = sorted(
        [build_asciidag(child, asciidag_nodes) for child in node.next],
        key=lambda n: n.item,
    )

    return asciidag_node
